@@ -19,11 +19,12 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <math.h>
 
 #include "pangocairo-private.h"
+#include "pango-glyph-item.h"
 
 typedef struct _PangoCairoRendererClass PangoCairoRendererClass;
 
@@ -37,7 +38,12 @@ struct _PangoCairoRenderer
 
   cairo_t *cr;
   gboolean do_path;
+  gboolean has_show_text_glyphs;
   double x_offset, y_offset;
+
+  /* house-keeping options */
+  gboolean is_cached_renderer;
+  gboolean cr_had_current_point;
 };
 
 struct _PangoCairoRendererClass
@@ -178,11 +184,7 @@ _pango_cairo_renderer_draw_box_glyph (PangoCairoRenderer *crenderer,
 				      double              cy,
 				      gboolean            invalid)
 {
-  double temp_x, temp_y;
-
   cairo_save (crenderer->cr);
-  cairo_get_current_point (crenderer->cr, &temp_x, &temp_y);
-
 
   _pango_cairo_renderer_draw_frame (crenderer,
 				    cx + 1.5,
@@ -192,7 +194,6 @@ _pango_cairo_renderer_draw_box_glyph (PangoCairoRenderer *crenderer,
 				    1.0,
 				    invalid);
 
-  cairo_move_to (crenderer->cr, temp_x, temp_y);
   cairo_restore (crenderer->cr);
 }
 
@@ -208,13 +209,11 @@ _pango_cairo_renderer_draw_unknown_glyph (PangoCairoRenderer *crenderer,
   int row, col;
   int rows, cols;
   char hexbox_string[2] = {0, 0};
-  double temp_x, temp_y;
   PangoCairoFontHexBoxInfo *hbi;
   gunichar ch;
   gboolean invalid_input;
 
   cairo_save (crenderer->cr);
-  cairo_get_current_point (crenderer->cr, &temp_x, &temp_y);
 
   ch = gi->glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
   invalid_input = G_UNLIKELY (gi->glyph == PANGO_GLYPH_INVALID_INPUT || ch > 0x10FFFF);
@@ -270,26 +269,33 @@ _pango_cairo_renderer_draw_unknown_glyph (PangoCairoRenderer *crenderer,
     }
 
 done:
-  cairo_move_to (crenderer->cr, temp_x, temp_y);
   cairo_restore (crenderer->cr);
 }
 
+#ifndef STACK_BUFFER_SIZE
+#define STACK_BUFFER_SIZE (512 * sizeof (int))
+#endif
+
+#define STACK_ARRAY_LENGTH(T) (STACK_BUFFER_SIZE / sizeof(T))
+
 static void
-pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
-				  PangoFont         *font,
-				  PangoGlyphString  *glyphs,
-				  int                x,
-				  int                y)
+pango_cairo_renderer_show_text_glyphs (PangoRenderer        *renderer,
+				       const char           *text,
+				       int                   text_len,
+				       PangoGlyphString     *glyphs,
+				       cairo_text_cluster_t *clusters,
+				       int                   num_clusters,
+				       gboolean              backward,
+				       PangoFont            *font,
+				       int                   x,
+				       int                   y)
 {
   PangoCairoRenderer *crenderer = (PangoCairoRenderer *) (renderer);
 
-  /* cairo_glyph_t is 24 bytes */
-#define MAX_STACK 40
-
   int i, count;
   int x_position = 0;
   cairo_glyph_t *cairo_glyphs;
-  cairo_glyph_t stack_glyphs[MAX_STACK];
+  cairo_glyph_t stack_glyphs[STACK_ARRAY_LENGTH (cairo_glyph_t)];
   double base_x = crenderer->x_offset + (double)x / PANGO_SCALE;
   double base_y = crenderer->y_offset + (double)y / PANGO_SCALE;
 
@@ -318,7 +324,7 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
       goto done;
     }
 
-  if (glyphs->num_glyphs > MAX_STACK)
+  if (glyphs->num_glyphs > (int) G_N_ELEMENTS (stack_glyphs))
     cairo_glyphs = g_new (cairo_glyph_t, glyphs->num_glyphs);
   else
     cairo_glyphs = stack_glyphs;
@@ -348,18 +354,118 @@ pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
       x_position += gi->geometry.width;
     }
 
-  if (crenderer->do_path)
+  if (G_UNLIKELY (crenderer->do_path))
     cairo_glyph_path (crenderer->cr, cairo_glyphs, count);
   else
-    cairo_show_glyphs (crenderer->cr, cairo_glyphs, count);
+    if (G_UNLIKELY (clusters))
+      cairo_show_text_glyphs (crenderer->cr,
+			      text, text_len,
+			      cairo_glyphs, count,
+			      clusters, num_clusters,
+			      backward);
+    else
+      cairo_show_glyphs (crenderer->cr, cairo_glyphs, count);
 
-  if (glyphs->num_glyphs > MAX_STACK)
+  if (cairo_glyphs != stack_glyphs)
     g_free (cairo_glyphs);
 
 done:
   cairo_restore (crenderer->cr);
+}
 
-#undef MAX_STACK
+static void
+pango_cairo_renderer_draw_glyphs (PangoRenderer     *renderer,
+				  PangoFont         *font,
+				  PangoGlyphString  *glyphs,
+				  int                x,
+				  int                y)
+{
+  pango_cairo_renderer_show_text_glyphs (renderer,
+					 NULL, 0,
+					 glyphs,
+					 NULL, 0,
+					 FALSE,
+					 font,
+					 x, y);
+}
+
+static void
+pango_cairo_renderer_draw_glyph_item (PangoRenderer     *renderer,
+				      const char        *text,
+				      PangoGlyphItem    *glyph_item,
+				      int                x,
+				      int                y)
+{
+  PangoCairoRenderer *crenderer = (PangoCairoRenderer *) (renderer);
+  PangoFont          *font      = glyph_item->item->analysis.font;
+  PangoGlyphString   *glyphs    = glyph_item->glyphs;
+  PangoItem          *item      = glyph_item->item;
+  gboolean            backward  = (item->analysis.level & 1) != 0;
+
+  PangoGlyphItemIter   iter;
+  cairo_text_cluster_t *cairo_clusters;
+  cairo_text_cluster_t stack_clusters[STACK_ARRAY_LENGTH (cairo_text_cluster_t)];
+  int num_clusters;
+
+  if (!crenderer->has_show_text_glyphs || crenderer->do_path)
+    {
+      pango_cairo_renderer_show_text_glyphs (renderer,
+					     NULL, 0,
+					     glyphs,
+					     NULL, 0,
+					     FALSE,
+					     font,
+					     x, y);
+      return;
+    }
+
+  if (glyphs->num_glyphs > (int) G_N_ELEMENTS (stack_clusters))
+    cairo_clusters = g_new (cairo_text_cluster_t, glyphs->num_glyphs);
+  else
+    cairo_clusters = stack_clusters;
+
+  num_clusters = 0;
+  if (pango_glyph_item_iter_init_start (&iter, glyph_item, text))
+    {
+      do {
+        int num_bytes, num_glyphs, i;
+
+        num_bytes  = iter.end_index - iter.start_index;
+        num_glyphs = backward ? iter.start_glyph - iter.end_glyph : iter.end_glyph - iter.start_glyph;
+
+	if (num_bytes < 1)
+	  g_warning ("pango_cairo_renderer_draw_glyph_item: bad cluster has num_bytess %d", num_bytes);
+	if (num_glyphs < 1)
+	  g_warning ("pango_cairo_renderer_draw_glyph_item: bad cluster has num_glyphs %d", num_glyphs);
+
+	/* Discount empty and unknown glyphs */
+	for (i = MIN (iter.start_glyph, iter.end_glyph+1);
+	     i < MAX (iter.start_glyph+1, iter.end_glyph);
+	     i++)
+	  {
+	    PangoGlyphInfo *gi = &glyphs->glyphs[i];
+
+	    if (gi->glyph == PANGO_GLYPH_EMPTY ||
+		gi->glyph & PANGO_GLYPH_UNKNOWN_FLAG)
+	      num_glyphs--;
+	  }
+
+        cairo_clusters[num_clusters].num_bytes  = num_bytes;
+        cairo_clusters[num_clusters].num_glyphs = num_glyphs;
+        num_clusters++;
+      } while (pango_glyph_item_iter_next_cluster (&iter));
+    }
+
+  pango_cairo_renderer_show_text_glyphs (renderer,
+					 text + item->offset, item->length,
+					 glyphs,
+					 cairo_clusters, num_clusters,
+					 backward,
+					 font,
+					 x, y);
+
+  if (cairo_clusters != stack_clusters)
+    g_free (cairo_clusters);
 }
 
 static void
@@ -548,6 +654,7 @@ pango_cairo_renderer_class_init (PangoCairoRendererClass *klass)
   PangoRendererClass *renderer_class = PANGO_RENDERER_CLASS (klass);
 
   renderer_class->draw_glyphs = pango_cairo_renderer_draw_glyphs;
+  renderer_class->draw_glyph_item = pango_cairo_renderer_draw_glyph_item;
   renderer_class->draw_rectangle = pango_cairo_renderer_draw_rectangle;
   renderer_class->draw_error_underline = pango_cairo_renderer_draw_error_underline;
   renderer_class->draw_shape = pango_cairo_renderer_draw_shape;
@@ -557,34 +664,36 @@ static PangoCairoRenderer *cached_renderer = NULL;
 G_LOCK_DEFINE_STATIC (cached_renderer);
 
 static PangoCairoRenderer *
-acquire_renderer (gboolean *free_renderer)
+acquire_renderer (void)
 {
   PangoCairoRenderer *renderer;
 
   if (G_LIKELY (G_TRYLOCK (cached_renderer)))
     {
       if (G_UNLIKELY (!cached_renderer))
-	cached_renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
+        {
+	  cached_renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
+	  cached_renderer->is_cached_renderer = TRUE;
+	}
 
       renderer = cached_renderer;
-      *free_renderer = FALSE;
     }
   else
     {
       renderer = g_object_new (PANGO_TYPE_CAIRO_RENDERER, NULL);
-      *free_renderer = TRUE;
     }
 
   return renderer;
 }
 
 static void
-release_renderer (PangoCairoRenderer *renderer, gboolean free_renderer)
+release_renderer (PangoCairoRenderer *renderer)
 {
-  if (G_LIKELY (!free_renderer))
+  if (G_LIKELY (renderer->is_cached_renderer))
     {
       renderer->cr = NULL;
       renderer->do_path = FALSE;
+      renderer->has_show_text_glyphs = FALSE;
       renderer->x_offset = 0.;
       renderer->y_offset = 0.;
 
@@ -594,6 +703,26 @@ release_renderer (PangoCairoRenderer *renderer, gboolean free_renderer)
     g_object_unref (renderer);
 }
 
+static void
+save_current_point (PangoCairoRenderer *renderer)
+{
+  renderer->cr_had_current_point = cairo_has_current_point (renderer->cr);
+  cairo_get_current_point (renderer->cr, &renderer->x_offset, &renderer->y_offset);
+
+  /* abuse save_current_point() to cache cairo_has_show_text_glyphs() result */
+  renderer->has_show_text_glyphs = cairo_has_show_text_glyphs (renderer->cr);
+}
+
+static void
+restore_current_point (PangoCairoRenderer *renderer)
+{
+  if (renderer->cr_had_current_point)
+    /* XXX should do cairo_set_current_point() when we have that function */
+    cairo_move_to (renderer->cr, renderer->x_offset, renderer->y_offset);
+  else
+    cairo_new_sub_path (renderer->cr);
+}
+
 
 /* convenience wrappers using the default renderer */
 
@@ -604,13 +733,12 @@ _pango_cairo_do_glyph_string (cairo_t          *cr,
 			      PangoGlyphString *glyphs,
 			      gboolean          do_path)
 {
-  gboolean free_renderer;
-  PangoCairoRenderer *crenderer = acquire_renderer (&free_renderer);
+  PangoCairoRenderer *crenderer = acquire_renderer ();
   PangoRenderer *renderer = (PangoRenderer *) crenderer;
 
   crenderer->cr = cr;
   crenderer->do_path = do_path;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
+  save_current_point (crenderer);
 
   if (!do_path)
     {
@@ -633,7 +761,48 @@ _pango_cairo_do_glyph_string (cairo_t          *cr,
       pango_renderer_deactivate (renderer);
     }
 
-  release_renderer (crenderer, free_renderer);
+  restore_current_point (crenderer);
+
+  release_renderer (crenderer);
+}
+
+static void
+_pango_cairo_do_glyph_item (cairo_t          *cr,
+			    const char       *text,
+			    PangoGlyphItem   *glyph_item,
+			    gboolean          do_path)
+{
+  PangoCairoRenderer *crenderer = acquire_renderer ();
+  PangoRenderer *renderer = (PangoRenderer *) crenderer;
+
+  crenderer->cr = cr;
+  crenderer->do_path = do_path;
+  save_current_point (crenderer);
+
+  if (!do_path)
+    {
+      /* unset all part colors, since when drawing just a glyph string,
+       * prepare_run() isn't called.
+       */
+
+      pango_renderer_activate (renderer);
+
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_FOREGROUND, NULL);
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_BACKGROUND, NULL);
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_UNDERLINE, NULL);
+      pango_renderer_set_color (renderer, PANGO_RENDER_PART_STRIKETHROUGH, NULL);
+    }
+
+  pango_renderer_draw_glyph_item (renderer, text, glyph_item, 0, 0);
+
+  if (!do_path)
+    {
+      pango_renderer_deactivate (renderer);
+    }
+
+  restore_current_point (crenderer);
+
+  release_renderer (crenderer);
 }
 
 static void
@@ -641,17 +810,18 @@ _pango_cairo_do_layout_line (cairo_t          *cr,
 			     PangoLayoutLine  *line,
 			     gboolean          do_path)
 {
-  gboolean free_renderer;
-  PangoCairoRenderer *crenderer = acquire_renderer (&free_renderer);
+  PangoCairoRenderer *crenderer = acquire_renderer ();
   PangoRenderer *renderer = (PangoRenderer *) crenderer;
 
   crenderer->cr = cr;
   crenderer->do_path = do_path;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
+  save_current_point (crenderer);
 
   pango_renderer_draw_layout_line (renderer, line, 0, 0);
 
-  release_renderer (crenderer, free_renderer);
+  restore_current_point (crenderer);
+
+  release_renderer (crenderer);
 }
 
 static void
@@ -659,17 +829,18 @@ _pango_cairo_do_layout (cairo_t     *cr,
 			PangoLayout *layout,
 			gboolean     do_path)
 {
-  gboolean free_renderer;
-  PangoCairoRenderer *crenderer = acquire_renderer (&free_renderer);
+  PangoCairoRenderer *crenderer = acquire_renderer ();
   PangoRenderer *renderer = (PangoRenderer *) crenderer;
 
   crenderer->cr = cr;
   crenderer->do_path = do_path;
-  cairo_get_current_point (cr, &crenderer->x_offset, &crenderer->y_offset);
+  save_current_point (crenderer);
 
   pango_renderer_draw_layout (renderer, layout, 0, 0);
 
-  release_renderer (crenderer, free_renderer);
+  restore_current_point (crenderer);
+
+  release_renderer (crenderer);
 }
 
 static void
@@ -722,6 +893,38 @@ pango_cairo_show_glyph_string (cairo_t          *cr,
   _pango_cairo_do_glyph_string (cr, font, glyphs, FALSE);
 }
 
+
+/**
+ * pango_cairo_show_glyph_item:
+ * @cr: a Cairo context
+ * @text: the UTF-8 text that @glyph_item refers to
+ * @glyph_item: a #PangoGlyphItem
+ *
+ * Draws the glyphs in @glyph_item in the specified cairo context,
+ * embedding the text associated with the glyphs in the output if the
+ * output format supports it (PDF for example), otherwise it acts
+ * similar to pango_cairo_show_glyph_string().
+ *
+ * The origin of the glyphs (the left edge of the baseline) will
+ * be drawn at the current point of the cairo context.
+ *
+ * Note that @text is the start of the text for layout, which is then
+ * indexed by <literal>@glyph_item->item->offset</literal>.
+ *
+ * Since: 1.22
+ **/
+void
+pango_cairo_show_glyph_item (cairo_t          *cr,
+			     const char       *text,
+			     PangoGlyphItem   *glyph_item)
+{
+  g_return_if_fail (cr != NULL);
+  g_return_if_fail (text != NULL);
+  g_return_if_fail (glyph_item != NULL);
+
+  _pango_cairo_do_glyph_item (cr, text, glyph_item, FALSE);
+}
+
 /**
  * pango_cairo_show_layout_line:
  * @cr: a Cairo context
