@@ -19,14 +19,13 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 #include <string.h>
 #include <stdlib.h>
 #include <math.h>
 
 #include <X11/Xlib.h>
 #include "pango-impl-utils.h"
-#include "modules.h"
 
 #undef PANGO_DISABLE_DEPRECATED
 
@@ -117,10 +116,7 @@ pango_x_find_subfont (PangoFont  *font,
   PangoXFont *xfont = (PangoXFont *)font;
 
   if (subfont_index < 1 || subfont_index > xfont->n_subfonts)
-    {
-      g_warning ("Invalid subfont %d", subfont_index);
-      return NULL;
-    }
+    return NULL;
 
   return xfont->subfonts[subfont_index-1];
 }
@@ -162,6 +158,30 @@ free_context_info (PangoXContextInfo *info)
   g_slice_free (PangoXContextInfo, info);
 }
 
+static PangoXContextInfo *
+get_context_info (PangoContext *context)
+{
+  PangoXContextInfo *info;
+  static quark = 0;
+
+  if (G_UNLIKELY (!quark))
+    quark = g_quark_from_static_string ("pango-x-info");
+
+  info =  g_object_get_qdata (G_OBJECT (context), quark);
+
+  if (G_UNLIKELY (!info))
+    {
+      info = g_slice_new (PangoXContextInfo);
+      info->get_gc_func = NULL;
+      info->free_gc_func = NULL;
+      g_object_set_qdata_full (G_OBJECT (context),
+			       quark,
+			       info, (GDestroyNotify)free_context_info);
+    }
+
+  return info;
+}
+
 /**
  * pango_x_get_context:
  * @display: an X display (As returned by XOpenDisplay().)
@@ -170,37 +190,14 @@ free_context_info (PangoXContextInfo *info)
  * given display.
  *
  * Return value: the new #PangoContext.
+ *
+ * Deprecated: 1.22: Use pango_x_font_map_for_display() followed by
+ * pango_font_map_create_context() instead.
  **/
 PangoContext *
 pango_x_get_context (Display *display)
 {
-  PangoContext *result;
-  PangoXContextInfo *info;
-  int i;
-  static gboolean registered_modules = FALSE;
-
-  g_return_val_if_fail (display != NULL, NULL);
-
-  if (!registered_modules)
-    {
-      registered_modules = TRUE;
-
-      for (i = 0; _pango_included_x_modules[i].list; i++)
-	pango_module_register (&_pango_included_x_modules[i]);
-    }
-
-  result = pango_context_new ();
-
-  info = g_slice_new (PangoXContextInfo);
-  info->get_gc_func = NULL;
-  info->free_gc_func = NULL;
-  g_object_set_qdata_full (G_OBJECT (result),
-			   g_quark_from_static_string ("pango-x-info"),
-			   info, (GDestroyNotify)free_context_info);
-
-  pango_context_set_font_map (result, pango_x_font_map_for_display (display));
-
-  return result;
+  return pango_font_map_create_context (pango_x_font_map_for_display (display));
 }
 
 /**
@@ -221,8 +218,7 @@ pango_x_context_set_funcs  (PangoContext     *context,
 
   g_return_if_fail (context != NULL);
 
-  info = g_object_get_qdata (G_OBJECT (context),
-			     g_quark_from_static_string ("pango-x-info"));
+  info = get_context_info (context);
 
   info->get_gc_func = get_gc_func;
   info->free_gc_func = free_gc_func;
@@ -302,8 +298,10 @@ pango_x_font_new (PangoFontMap *fontmap, const char *spec, int size)
 
   result = g_object_new (PANGO_TYPE_X_FONT, NULL);
 
+  g_assert (result->fontmap == NULL);
   result->fontmap = fontmap;
-  g_object_ref (fontmap);
+  g_object_add_weak_pointer (G_OBJECT (result->fontmap), (gpointer *) (gpointer) &result->fontmap);
+
   result->display = pango_x_fontmap_get_display (fontmap);
 
   result->fonts = g_strsplit(spec, ",", -1);
@@ -326,7 +324,7 @@ pango_x_font_new (PangoFontMap *fontmap, const char *spec, int size)
  * work correctly in all circumstances. Use of this function should
  * be avoided.
  *
- * Returns a new #PangoFont.
+ * Returns: a new #PangoFont.
  */
 PangoFont *
 pango_x_load_font (Display    *display,
@@ -411,7 +409,7 @@ pango_x_render  (Display           *display,
 	    glyph_y >= -16384 && glyph_y <= 32767))
 	goto next_glyph;
 
-      if ((glyph & PANGO_GLYPH_UNKNOWN_FLAG) == 0)
+      if (G_LIKELY ((glyph & PANGO_GLYPH_UNKNOWN_FLAG) == 0))
 	{
 	  guint16 index = PANGO_X_GLYPH_INDEX (glyph);
 	  guint16 subfont_index = PANGO_X_GLYPH_SUBFONT (glyph);
@@ -455,6 +453,7 @@ pango_x_render  (Display           *display,
 	  int baseline;
 	  int stroke_thick;
 	  gunichar wc;
+	  gboolean invalid_input;
 
 	unknown_glyph:
 	  FLUSH;
@@ -479,12 +478,13 @@ pango_x_render  (Display           *display,
 	  x2 = x1 + PANGO_PIXELS (glyphs->glyphs[i].geometry.width);
 
 	  baseline = glyph_y;
-	  stroke_thick = MAX ((int) (0.5 + 0.075 * (y2 - y1)), 1);
+	  stroke_thick = MAX ((int) (0.5 + 0.025 * (y2 - y1)), 1);
 
 	  if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
-	    wc = glyph & (~PANGO_GLYPH_UNKNOWN_FLAG);
+	    wc = glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
 	  else
 	    wc = 0;
+	  invalid_input = glyph == PANGO_GLYPH_INVALID_INPUT || wc > 0x10FFFF;
 
 	  switch (wc)
 	    {
@@ -544,7 +544,7 @@ pango_x_render  (Display           *display,
 	    default:
 	      {
 		/* Perhaps we should draw the box-with-numbers as in the
-		 * Xft backend, though we have no guarantee of having
+		 * other backends, though we have no guarantee of having
 		 * an appropriate size of font. Right now, we just
 		 * draw an empty box. (To draw the box-with-numbers.
 		 * the backends would have to be changed to use
@@ -572,6 +572,15 @@ pango_x_render  (Display           *display,
 		XFillRectangle (display, d, gc,
 				x1, y2 - stroke_thick,
 				x2 - x1, stroke_thick);
+		if (invalid_input)
+		  {
+		    XDrawLine (display, d, gc,
+			       x1, y1,
+			       x2-1, y2-1);
+		    XDrawLine (display, d, gc,
+			       x2-1, y1,
+			       x1, y2-1);
+		  }
 
 		break;
 	      }
@@ -761,7 +770,7 @@ itemize_string_foreach (PangoFont     *font,
 	  (finished ||
 	   (shaper != last_shaper || last_level != embedding_levels[i])))
 	{
-	  PangoAnalysis analysis = {0};
+	  PangoAnalysis analysis = { NULL };
 	  int j;
 
 	  analysis.shape_engine = last_shaper;
@@ -1311,7 +1320,9 @@ pango_x_font_finalize (GObject *object)
   if (xfont->xface)
     pango_x_face_remove (xfont->xface, (PangoFont *)xfont);
 
-  g_object_unref (xfont->fontmap);
+  g_assert (xfont->fontmap != NULL);
+  g_object_remove_weak_pointer (G_OBJECT (xfont->fontmap), (gpointer *) (gpointer) &xfont->fontmap);
+  xfont->fontmap = NULL;
 
   g_strfreev (xfont->fonts);
 
@@ -1489,9 +1500,7 @@ pango_x_render_layout_line (Display          *display,
   PangoRectangle logical_rect;
   PangoRectangle ink_rect;
   PangoContext *context = pango_layout_get_context (line->layout);
-  PangoXContextInfo *info =
-    g_object_get_qdata (G_OBJECT (context),
-			g_quark_from_static_string ("pango-x-info"));
+  PangoXContextInfo *info = get_context_info (context);
 
   int x_off = 0;
 
